/*
 * simulate 
 * argv[1] = <spring constant>
 * argv[2] = <rest length> 
 * argv[3] = <mass>
 * argv[4-6] = <anchor x, y, z>
 * argv[7-9] = <initial pos>
 * argv[10-12] = <initial velocity>
 * argv[13-15] = <grav accel>
 * argv[16] = <efficiency>
 * argv[17] = <start frame> (corresponds to initial pos/vel
 * argv[18] = <end frame>
 * argv[19] = <discritization> (the number of subframes to use for calculation)
 * argv[20] = <fps>
 * argv[21] = <save file>
 *
 */

#include <math.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>

/* **** VECTORS **** */
typedef struct {
	float x, y, z;
} vector_t;

/* Magnitude and dot product are simple enough to be macros */
#define dot(a,b)	((a).x * (b).x + (a).y * (b).y + (a).z * (b).z)
#define vector_mag(v)	sqrt(dot(v,v))

inline vector_t vector_add(vector_t v1, vector_t v2) {
	v1.x += v2.x;
	v1.y += v2.y;
	v1.z += v2.z;

	return v1;
}

inline vector_t vector_sub(vector_t v1, vector_t v2) {
	v1.x -= v2.x;
	v1.y -= v2.y;
	v1.z -= v2.z;

	return v1;
}

inline vector_t vector_scale(float s, vector_t v) {
	v.x *= s;
	v.y *= s;
	v.z *= s;

	return v;
}

inline vector_t unit_vector(vector_t v) {
	return vector_scale(1.0 / vector_mag(v), v);
}

/* What is thr force at the end point */
vector_t spring_force(vector_t anchor, vector_t end, float k, float lRest) {
	vector_t deltaP = vector_sub(end, anchor);

	/* how much are we stretched? */
	float l = vector_mag(deltaP);

	/* So what is the force */
	float force = (lRest - l) * k;

	/* And return it in vector form */
	return vector_scale(force, unit_vector(deltaP));
}

/*
 * simulate 
 * argv[1] = <spring constant>
 * argv[2] = <rest length> 
 * argv[3] = <mass>
 * argv[4-6] = <anchor x, y, z>
 * argv[7-9] = <initial pos>
 * argv[10-12] = <initial velocity>
 * argv[13-15] = <grav accel>
 * argv[16] = <efficiency>
 * argv[17] = <start frame> (corresponds to initial pos/vel
 * argv[18] = <end frame>
 * argv[19] = <discritization> (the number of subframes to use for calculation)
 * argv[20] = <fps>
 * argv[21] = <save file>
 *
 */

int main(int argc, char **argv) {
	/* Assume we're called intelligently */

	/* Passed Arguments */
	float k = atof(argv[1]);
	float lRest = atof(argv[2]);
	float mass = atof(argv[3]);
	vector_t anchor = {atof(argv[4]), atof(argv[5]), atof(argv[6])};
	vector_t pInitial = {atof(argv[7]), atof(argv[8]), atof(argv[9])};
	vector_t vInitial = {atof(argv[10]), atof(argv[11]), atof(argv[12])};
	vector_t aGrav = {atof(argv[13]), atof(argv[14]), atof(argv[15])};
	float efficiency = atof(argv[16]);
	int32_t startFrame= atoi(argv[17]);
	int32_t endFrame = atoi(argv[18]);
	uint32_t disc = atoi(argv[19]);
	float framerate = atof(argv[20]);
	const char *cacheFile = argv[21];

	/* Variables Used During Simulation */

	/* F = ma */
	vector_t fGrav = vector_scale(mass, aGrav);
	vector_t fSpring;
	vector_t fTotal;

	/* Current time in seconds.  0 corresponds to startFrame */
	float time = 0.0;

	/* Elapsed time since the last frame */
	float delta_t;

	/* position, velocity, and accel */
	vector_t pos=pInitial;
	vector_t vel=vInitial;
	vector_t accel;

	/* The new framerate with our higher discritization */
	float discrate = framerate * disc;

	/* This is the current frame in our internal reference (at discrate, starting at 0) */
	uint32_t frame = 0;
	uint32_t lastFrame = (endFrame-startFrame) * disc;

	/* efficiency is the ammount of energy conserved over one second
	 * disceff is the ammount of energy conserved over one frame
	 *
	 * disceff^discrate = efficiency
	 * => discrate * log2(disceff) = log(efficiency)
	 * => disceff = 2 ^ (log2(efficiency) / discrate)
	 */
	float disceff = pow(2, log2(efficiency) / discrate);
	float energy;

	/* Our output file handle */
	FILE *fp = fopen(cacheFile, "w");

	/* We want to print the initial position */
	fprintf(fp, "%f %f %f\n", pos.x, pos.y, pos.z);

	/* Now, calculate subsequent locations */
	for(frame=1; frame <= lastFrame; frame++) {

		/* How much time is elapsing during this itteration from the previous frame to the current one? */
		delta_t = (frame / discrate) - time;
		time = frame / discrate;

		/* Get our force */
		fSpring = spring_force(anchor, pos, k, lRest);
		fTotal = vector_add(fSpring, fGrav);

		/* Modify pos by old v */
		pos = vector_add(pos, vector_scale(delta_t, vel));

		/* Modify vel by new accel */
		/* F = ma */
		accel = vector_scale(1.0 / mass, fTotal);
		vel = vector_add(vel, vector_scale(delta_t, accel));

		/* The percentage of energy conserved is equal to disceff */
		energy = vector_mag(vel);
		energy *= energy;  /* Current energy : K proportional to v^2 */
		energy *= disceff; /* Factor in energy loss */
		vel = vector_scale(sqrt(energy), unit_vector(vel)); /* Rescale the vector */

		/* Report the position if this is an actual frame */		
		if( frame % disc == 0 ) {
			fprintf(fp, "%f %f %f\n", pos.x, pos.y, pos.z);
		}

		fprintf(stderr, "F[%u.%u]: t=%.2f p=<%.2f, %.2f, %.2f> v=<%.2f, %.2f, %.2f> a=<%.2f, %.2f, %.2f>\n", frame / disc, frame % disc, time, pos.x, pos.y, pos.z, vel.x, vel.y, vel.z, accel.x, accel.y, accel.z);
	}

	return 0;
}
